#ifndef __PNGENCOD_H
#define __PNGENCOD_H
//
// Copyright (c) 1997,1998 Colosseum Builders, Inc.
// All rights reserved.
//
// Colosseum Builders, Inc. makes no warranty, expressed or implied
// with regards to this software. It is provided as is.
//
// See the README.TXT file that came with this software for restrictions
// on the use and redistribution of this file or send E-mail to
// info@colosseumbuilders.com
//

//
//  Title: PNG Encoder Class Definition
//
//  Author:  John M. Miano  miano@colosseumbuilders.com
//
//  Description:
//
//    This class implements PNG encoding.
//

#include "png.h"
#include "bitimage.h"

class PngHuffmanEncoder ;

class PngEncoder : public BitmapImageEncoder
{
public:
  PngEncoder () ;
  PngEncoder (const PngEncoder &) ;
  virtual ~PngEncoder () ;

  PngEncoder &operator=(const PngEncoder &) ;

  void WriteImage (std::ostream &, BitmapImage &) ;

  // Property functions for predefined tEXt strings.
  void SetTitle (const std::string &) ;
  std::string GetTitle () const ;

  void SetAuthor (const std::string &) ;
  std::string GetAuthor () const ;

  void SetDescription (const std::string &) ;
  std::string GetDescription () const ;

  void SetCopyright (const std::string &) ;
  std::string GetCopyright () const ;

  void SetSoftware (const std::string &) ;
  std::string GetSoftware () const ;

  void SetDisclaimer (const std::string &) ;
  std::string GetDisclaimer () const ;

  void SetWarning (const std::string &) ;
  std::string GetWarning () const ;

  void SetSource (const std::string &) ;
  std::string GetSource () const ;

  void SetComment (const std::string &) ;
  std::string GetComment () const ;

  unsigned long GetBlockSize () const ;
  void SetBlockSize (unsigned long) ;

  enum CompressionLevel
  {
    FastestCompression = 0,
    FastCompression = 1,
    DefaultCompression = 2,
    MaximumCompression = 3,
  } ;

  void SetCompressionLevel (CompressionLevel) ;  //  
  CompressionLevel GetCompressionLevel () const ;
  bool compressionVertical, compressionGorizontal, compressionAverage, compressionDetect;
  bool trebaOptimizeLZBlock;     //   LZ- ,kjre
  void SetCompressionVertical();
  void SetCompressionGorizontal();
  void SetCompressionAverage();
  void SetCompressionDetect();
  UBYTE1 cm11, cm12, cm21, cm22; //     

protected:
  void Initialize () ;
  void DoCopy (const PngEncoder &) ;

private:
  unsigned long PngWindowSize; // -  PNG-

  struct HashEntry
  {
    unsigned long index;
    HashEntry *next ;
  } ;

  UBYTE1 *imageData;         //   
  typedef HashEntry *PHashEntry;
  HashEntry *hash_values;
  PHashEntry *hash_table;
  unsigned long *countHashValue;

  UBYTE1 *vimageData;         //    
  HashEntry *vhash_values;
  PHashEntry *vhash_table;
  unsigned long *vcountHashValue;

  UBYTE1 *gimageData;         //    
  HashEntry *ghash_values;
  PHashEntry *ghash_table;
  unsigned long *gcountHashValue;

  UBYTE1 *aimageData;         //    
  HashEntry *ahash_values;
  PHashEntry *ahash_table;
  unsigned long *acountHashValue;

  UBYTE1 *dimageData;         //    
  HashEntry *dhash_values;
  PHashEntry *dhash_table;
  unsigned long *dcountHashValue;

  void initCM();   //      
  void doneCM();    //    
  void writeCM();  //     
  void writeNotCP();//      

  UBYTE1 countComponent;     //   
  UBYTE4 currentPozImage, countByteImage, countPointImage;
  UBYTE4 * nextPozStartBlockFilter; //       
  UBYTE2 indexBlockFilter;
  void compareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void vcompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void gcompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void acompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   
  void dcompareStartPoslidovnist(long adres1, long adres2, unsigned int &countRivni, unsigned int bestLength); //   

  void LongestMatch (unsigned int &length,
                     unsigned int &offset) ;
  void VLongestMatch(unsigned int &length,
                     unsigned int &offset) ;
  void GLongestMatch(unsigned int &length,
                     unsigned int &offset) ;
  void ALongestMatch(unsigned int &length,
                     unsigned int &offset) ;
  void DLongestMatch(unsigned int &length,
                     unsigned int &offset) ;
  unsigned int HashValue (unsigned long) ;
  unsigned int VHashValue (unsigned long) ;
  unsigned int GHashValue (unsigned long) ;
  unsigned int AHashValue (unsigned long) ;
  unsigned int DHashValue (unsigned long) ;

  void FreePozHashEntry(unsigned long entry);
  void VFreePozHashEntry(unsigned long entry);
  void GFreePozHashEntry(unsigned long entry);
  void AFreePozHashEntry(unsigned long entry);
  void DFreePozHashEntry(unsigned long entry);

  void MoveHashEntry (unsigned long entry, unsigned int hashvalue) ;
  void VMoveHashEntry (unsigned long entry, unsigned int hashvalue) ;
  void GMoveHashEntry (unsigned long entry, unsigned int hashvalue) ;
  void AMoveHashEntry (unsigned long entry, unsigned int hashvalue) ;
  void DMoveHashEntry (unsigned long entry, unsigned int hashvalue) ;

  BitmapImage *current_image ;
  std::ostream *output_stream ;

  PngHuffmanEncoder *distance_table ;
  PngHuffmanEncoder *length_table ;
  PngHuffmanEncoder *length_length_table ;

  typedef void (PngEncoder::*LENGTHFUNCTION) (unsigned int index,
                                              unsigned int code,
                                              unsigned int length,
                                              unsigned int extra) ;
  void GatherLengthCounts (unsigned int,
                           unsigned int code,
                           unsigned int,
                           unsigned int) ;
  void OutputLengthCounts (unsigned int index,
                           unsigned int code,
                           unsigned int extra,
                           unsigned int value) ;

  void FindLengthCodes (LENGTHFUNCTION, UBYTE1 lengths [], unsigned int count) ;

  bool ProcessImageData () ;

  UBYTE1 *output_buffer ;
  UBYTE4 chunk_type ;
  UBYTE4 byte_position ;
  unsigned int bit_position ;
  void StartChunk (const char type [5]) ;
  void WriteCurrentChunk () ;
  void OutputBlock (const void *data, unsigned int length) ;
  void OutputByte (UBYTE1) ;
  void OutputDataBits (unsigned int data, unsigned int length) ;
  void FlushBitBuffer () ;
  void OutputAdler () ;

  void FreeBuffers () ;

  int row_width ; //  unsigned
  void WriteText (const std::string &keyword, const std::string &value) ;
  void WriteTextBlocks () ;

  void CallProgressFunction (unsigned int) ;

  std::string title_string ;
  std::string author_string ;
  std::string description_string ;
  std::string copyright_string ;
  std::string software_string ;
  std::string disclaimer_string ;
  std::string warning_string ;
  std::string source_string ;
  std::string comment_string ;

  void DoWrite () ;

  void InitializeHashTable () ;
  void InitializeLZWindow () ;
  void OutputDeflateHeader (bool lastblock) ;
  void OutputZLibHeader () ;
  void WriteImageData () ;

  CompressionLevel compression_level;
  unsigned int search_limit;
  UBYTE4 adler_value ;
  unsigned int image_row ;
  unsigned int image_col ;
  bool image_end ;

  UBYTE2 *block_buffer ;
  PngHuffmanEncoder *length_literal_table;   // ,   
  PngHuffmanEncoder *distance_literal_table; // ,   
  UBYTE1 * trebaZaminaBuffer;  // LZ-   
  UBYTE4 countAnalizZamina;    //     
  UBYTE4 * pozAnalizZamina;    //     
  UBYTE4 * pozImageBuffer;     //     
  unsigned int block_buffer_size ;
  unsigned int block_buffer_count ;

  void OutputBlockData () ;
} ;

inline void PngEncoder::SetTitle (const std::string &value)
{
  title_string = value ;
  return ;
}

inline std::string PngEncoder::GetTitle () const
{
  return title_string ;
}

inline void PngEncoder::SetAuthor (const std::string &value)
{
  author_string = value ;
  return ;
}

inline std::string PngEncoder::GetAuthor () const
{
  return author_string ;
}

inline void PngEncoder::SetDescription (const std::string &value)
{
  description_string = value ;
  return ;
}

inline std::string PngEncoder::GetDescription () const
{
  return description_string ;
}

inline void PngEncoder::SetCopyright (const std::string &value)
{
  copyright_string = value ;
  return ;
}

inline std::string PngEncoder::GetCopyright () const
{
  return copyright_string ;
}

inline void PngEncoder::SetSoftware (const std::string &value)
{
  software_string = value ;
  return ;
}

inline std::string PngEncoder::GetSoftware () const
{
  return software_string ;
}

inline void PngEncoder::SetDisclaimer (const std::string &value)
{
  disclaimer_string = value ;
  return ;
}

inline std::string PngEncoder::GetDisclaimer () const
{
  return disclaimer_string ;
}

inline void PngEncoder::SetWarning (const std::string &value)
{
  warning_string = value ;
  return ;
}

inline std::string PngEncoder::GetWarning () const
{
  return warning_string ;
}

inline void PngEncoder::SetSource (const std::string &value)
{
  source_string = value ;
  return ;
}

inline std::string PngEncoder::GetSource () const
{
  return source_string ;
}

inline void PngEncoder::SetComment (const std::string &value)
{
  comment_string = value ;
  return ;
}

inline std::string PngEncoder::GetComment () const
{
  return comment_string ;
}

inline unsigned long PngEncoder::GetBlockSize () const
{
  return block_buffer_size ;
}

inline void PngEncoder::SetBlockSize (unsigned long value)
{
  if (value < 500)
    block_buffer_size = 500 ;
  else
    block_buffer_size = value ;
  return ;
}

#endif


